# psim を利用するためのおまじない -- ここから
import winreg, sys, os
reg = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, "SOFTWARE\\Wow6432Node\\Mathematical Systems Inc.\\S-Quattro Simulation System\\INFO")
rootdir = winreg.QueryValueEx(reg, "ROOTDIR")[0]
os.environ["PATH"] = rootdir + ";" + os.environ["PATH"]
sys.path.insert(0, os.path.join(rootdir, "lib3"))
# psim を利用するためのおまじない -- ここまで

from psim import *
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import pdb

def sim(ns):
    initialize()

    lmd = 1.0 / 60 # 平均到着率
    r = [0.6, 0.1, 0.3] # 住民票、戸籍、印鑑証明の割合
    m = [3 * 60, 4 * 60, 2 * 60] # 住民票、戸籍、印鑑証明のサービス時間

    # 到着間隔を生成する乱数発生器
    g1 = exponentialDistribution(1 / lmd)
    # サービス種別(住民票、戸籍、印鑑証明)を返す乱数発生器
    gs = empiricalDistribution([(v, i) for i, v in enumerate(r)])
    # 各サービス時間を生成する乱数発生器
    gt = [exponentialDistribution(m0) for m0 in m]
    # 窓口を示すファシリティ資源
    f = Facility(ns, monitor = True)
    # 待ち時間を記録するモニター
    monitor = Monitor(["waitingTime"], ["f"], name = "wt")

    # 到着プロセス
    def arrive():
        while True:
            # 次の到着まで待つ
            yield pause(next(g1))
            # サービス種別を決定
            s = next(gs)
            # サービスプロセスを開始する
            # - この処理自体は非同期なので、一瞬で完了する
            yield subactivate(service)(s)

    # サービスプロセス
    def service(s):
        t0 = now()
        # 窓口の待ち行列に並ぶ
        # - 窓口を示すファシリティ資源の利用を予約する
        # - 既に利用者がいる場合は、利用可能になるまで待ち受ける
        lock = (yield f.request(name = "lock"))["lock"]
        # 窓口サービスを開始する
        # - 窓口を示すファシリティ資源の利用権を取得した
        # - 利用権を開放するまで他の利用者は利用できない
        t1 = now()
        # 待ち時間を記録
        monitor.observe(t1 - t0)
        # 窓口を利用する
        yield pause(next(gt[s]))
        # 窓口を示すファシリティ資源の利用権を開放する
        lock.release()

    # 到着プロセスの開始を登録する
    activate(arrive)()

    # シミュレーション開始
    start(until = 100000)

    # M/M/1 の理論値
    #rho = lmd / mu # 平均利用率
    #Q = rho / (1.0 - rho) # 平均待ち行列
    #W = Q / mu # 平均待ち時間

    waitingTime = monitor["waitingTime"].mean()
    # 平均待ち行列(窓口利用者をカウントしない)
    queue0 = f.monitor["要求待ち"].mean()
    # 平均システム内客数(窓口利用者をカウントする)
    queue = (f.monitor["要求待ち"] + f.monitor["バッファ"]).mean()
    return waitingTime, queue

result = []
for ns in range(1, 5):    
    waitingTime, queue = sim(ns)
    result.append([ns, waitingTime, queue])

df = pd.DataFrame(result, columns = ["窓口数", "待ち時間", "客数"])
print(df)

plt.rcParams['font.sans-serif'].insert(0, "Meiryo")
fig, ax = plt.subplots(1, 2)
ax[0].bar(df["窓口数"], df["待ち時間"])
ax[0].set_ylabel("待ち時間(秒)")
ax[0].set_xlabel("窓口数")
ax[1].bar(df["窓口数"], df["客数"])
ax[1].set_ylabel("客数(人)")
ax[1].set_xlabel("窓口数")
plt.tight_layout()
#plt.show()
plt.savefig("graph.png")
